@@ -27,6 +27,7 @@ import traceback
 import sqlite3
 import copy
 import collections
+import portalocker
 import numpy as np
 from datetime import datetime, timedelta
 from json import dumps, loads
@@ -299,52 +300,30 @@ class JobData(object):
         Returns:
             int: queueing time
         """
-        # if self.job_name == 'a3bh_18500101_fc1_138_SIM':
-        #     print("Package {} {} {}".format(
-        #         self.job_name, self.rowtype, self._start))
-        # We consider the highest queue time in the package as the queue time of the whole package.
         max_queue = queue = 0
         job_name_max_queue = None
-        # print(job_data_in_package)
-        # print("Run {} Job {}".format(self.run_id, self.job_name))
-        # print("Raw queue {}".format(self._start - self._submit))
+
         if job_data_in_package and len(job_data_in_package) > 0:
             # Only consider those jobs with starting time less than the start time of the job minus 20 seconds.
 
             jobs_times = [job_data_in_package[key]
                           for key in job_data_in_package if job_data_in_package[key][3] < (self._start - 20)]
-            # if self.job_name == 'a3bh_18500101_fc1_138_SIM':
-            #     for package in job_data_in_package:
-            #         print(job_data_in_package[package])
-            # print(jobs_times)
+
             if jobs_times and len(jobs_times) > 0:
+                # There are previous jobs
+                # Sort by Queuing Time from Highest to Lowest
                 jobs_times.sort(key=lambda a: a[0], reverse=True)
-                # print(jobs_times[0])
+                # Select the maximum queue time
                 max_queue, _, job_name_max_queue, start, finish = jobs_times[0]
+                # Add the running time to the max queue time
                 max_queue += (finish - start) if finish > start else 0
-                # print(jobs_times)
-                # if self.job_name == 'a3bh_18500101_fc1_138_SIM':
-                #     print("Current {}".format(self.start - self.submit))
-                #     print("Max Queue {}".format(max_queue))
-                #     print("Job Name Max {}".format(job_name_max_queue))
 
         if self.status in ["SUBMITTED", "QUEUING", "RUNNING", "COMPLETED", "HELD", "PREPARED", "FAILED"]:
-            # parents_sum = 0
-            # print("{} {}".format(self.job_name, self.rowtype))
-            # if self.rowtype > 2:
-            #     # if job_name_max_queue and self.job_name == job_name_max_queue:
-            #     #     # path = []
-            #     #     queue = int((self.start if self.start >
-            #     #                  0 else time.time()) - self.submit)
-            #     # else:
-            #     queue = int((self.start if self.start >
-            #                      0 else time.time()) - self.submit) - int(max_queue)
-            # else:
-
+            # Substract the total time from the max_queue job in the package
+            # This adjustment should cover most of the wrapper types.
+            # TODO: Test this mechanism against all wrapper types
             queue = int((self.start if self.start >
                          0 else time.time()) - self.submit) - int(max_queue)
-            # if self.job_name == 'a3bh_18500101_fc1_138_SIM':
-            #     print("Queue {}".format(queue))
             if queue > 0:
                 return queue
         return 0
@@ -541,6 +520,216 @@ class MainDataBase():
             return None
 
 
+class ExperimentGraphDrawing(MainDataBase):
+    def __init__(self, expid):
+        """
+        Sets and validates graph drawing.
+        :param expid: Name of experiment  
+        :type expid: str  
+        :param allJobs: list of all jobs objects (usually from job_list)  
+        :type allJobs: list()
+        """
+        MainDataBase.__init__(self, expid)
+        BasicConfig.read()
+        self.expid = expid
+        self.folder_path = BasicConfig.GRAPHDATA_DIR
+        self.database_path = os.path.join(
+            self.folder_path, "graph_data_" + str(expid) + ".db")
+        self.create_table_query = textwrap.dedent(
+            '''CREATE TABLE
+        IF NOT EXISTS experiment_graph_draw (        
+        id INTEGER PRIMARY KEY,
+        job_name text NOT NULL,
+        x INTEGER NOT NULL,
+        y INTEGER NOT NULL        
+        );''')
+
+        if not os.path.exists(self.database_path):
+            self.conn = self.create_connection(self.database_path)
+            self.create_table()
+        else:
+            self.conn = self.create_connection(self.database_path)
+        self.lock_name = "calculation_in_progress.lock"
+        self.current_position_dictionary = None
+        self.current_jobs_set = set()
+        self.coordinates = list()
+        self.set_current_position()
+        self.should_update = False
+        self.locked = False
+        self.test_locked()
+
+    def test_locked(self):
+        self.locked = True
+        try:
+            with portalocker.Lock(os.path.join(self.folder_path, self.lock_name), timeout=1) as fh:
+                self.locked = False
+                fh.flush()
+                os.fsync(fh.fileno())
+        except portalocker.AlreadyLocked:
+            print("It is locked")
+            self.locked = True
+        except Exception as exp:
+            self.locked = True
+
+    def get_validated_data(self, allJobs):
+        """
+        Validates if should update current graph drawing.  
+        :return: None if graph drawing should be updated, otherwise, it returns the position data.  
+        :rype: None or dict()
+        """
+        job_names = {job.name for job in allJobs}
+        # Validating content
+        difference = job_names - self.current_jobs_set
+        if difference and len(difference) > 0:
+            # Intersection found. Graph Drawing database needs to be updated
+            self.should_update = True
+            # Clear database
+        return None if self.should_update == True else self.current_position_dictionary
+
+    def calculate_drawing(self, allJobs, independent=False, num_chunks=48):
+        """
+        Called in a thread.
+        :param allJobs: list of jobs (usually from job_list object)  
+        :type allJobs: list()  
+        :return: Last row Id  
+        :rtype: int 
+        """
+        lock_name = "calculation_{}_in_progress.lock".format(
+            self.expid) if independent == True else self.lock_name
+        lock_path_file = os.path.join(self.folder_path, lock_name)
+        try:
+            with portalocker.Lock(lock_path_file, timeout=1) as fh:
+                self.conn = self.create_connection(self.database_path)
+                monitor = Monitor()
+                graph = monitor.create_tree_list(
+                    self.expid, allJobs, None, dict(), False)
+                if len(allJobs) > 1000:
+                    # Logic: Start with 48 as acceptable number of chunks for Gmaxiter = 100
+                    # Minimum Gmaxiter will be 10
+                    maxiter = max(10, 148 - num_chunks)
+                    print("Experiment {} num_chunk {} maxiter {}".format(
+                        self.expid, num_chunks, maxiter))
+                    result = graph.create(
+                        ['dot', '-Gnslimit=2', '-Gnslimit1=2', '-Gmaxiter={}'.format(maxiter), '-Gsplines=none', '-v'], format="plain")
+                else:
+                    result = graph.create('dot', format="plain")
+                for u in result.split("\n"):
+                    splitList = u.split(" ")
+                    if len(splitList) > 1 and splitList[0] == "node":
+                        self.coordinates.append((splitList[1], int(
+                            float(splitList[2]) * 90), int(float(splitList[3]) * -90)))
+                        # self.coordinates[splitList[1]] = (
+                        #     int(float(splitList[2]) * 90), int(float(splitList[3]) * -90))
+                self.insert_coordinates()
+                fh.flush()
+                os.fsync(fh.fileno())
+            os.remove(lock_path_file)
+            return self.get_validated_data(allJobs)
+        except portalocker.AlreadyLocked:
+            message = "Already calculating graph drawing."
+            print(message)
+            return None
+        except Exception as exp:
+            print(traceback.format_exc())
+            os.remove(lock_path_file)
+            print("Exception while calculating coordinates {}".format(str(exp)))
+            return None
+
+    def insert_coordinates(self):
+        """
+        Prepares and inserts new coordinates.
+        """
+        try:
+            # Start by clearing database
+            self._clear_graph_database()
+            result = None
+            if self.coordinates and len(self.coordinates) > 0:
+                result = self._insert_many_graph_coordinates(self.coordinates)
+                return result
+            return None
+        except Exception as exp:
+            print(str(exp))
+            return None
+
+    def set_current_position(self):
+        """
+        Sets all registers in the proper variables.  
+        current_position_dictionary: JobName -> (x, y)  
+        current_jobs_set: JobName
+        """
+        current_table = self._get_current_position()
+        if current_table and len(current_table) > 0:
+            self.current_position_dictionary = {
+                row[1]: (row[2], row[3]) for row in current_table}
+            self.current_jobs_set = set(
+                self.current_position_dictionary.keys())
+
+    def _get_current_position(self):
+        """
+        Get all registers from experiment_graph_draw.\n
+        :return: row content: id, job_name, x, y  
+        :rtype: 4-tuple (int, str, int, int)
+        """
+        try:
+            if self.conn:
+                # conn = create_connection(DB_FILE_AS_TIMES)
+                self.conn.text_factory = str
+                cur = self.conn.cursor()
+                cur.execute(
+                    "SELECT id, job_name, x, y FROM experiment_graph_draw")
+                rows = cur.fetchall()
+                return rows
+            return None
+        except Exception as exp:
+            print(traceback.format_exc())
+            print(str(exp))
+            return None
+
+    def _insert_many_graph_coordinates(self, values):
+        """
+        Create many graph coordinates
+        :param conn:
+        :param details:
+        :return:
+        """
+        try:
+            if self.conn:
+                # exp_id = self._get_id_db()
+                # conn = create_connection(DB_FILE_AS_TIMES)
+                # creation_date = datetime.today().strftime('%Y-%m-%d-%H:%M:%S')
+                sql = ''' INSERT INTO experiment_graph_draw(job_name, x, y) VALUES(?,?,?) '''
+                # print(row_content)
+                cur = self.conn.cursor()
+                cur.executemany(sql, values)
+                # print(cur)
+                self.conn.commit()
+                return cur.lastrowid
+        except Exception as exp:
+            print(traceback.format_exc())
+            Log.warning(
+                "Error on Insert many graph drawing : {}".format(str(exp)))
+            return None
+
+    def _clear_graph_database(self):
+        """
+        Clear all content from graph drawing database
+        """
+        try:
+            if self.conn:
+                # conn = create_connection(DB_FILE_AS_TIMES)
+                # modified_date = datetime.today().strftime('%Y-%m-%d-%H:%M:%S')
+                sql = ''' DELETE FROM experiment_graph_draw '''
+                cur = self.conn.cursor()
+                cur.execute(sql, )
+                self.conn.commit()
+                return True
+            return False
+        except Exception as exp:
+            print(traceback.format_exc())
+            print("Error on Database clear: {}".format(str(exp)))
+            return False
+
+
 class ExperimentStatus(MainDataBase):
     def __init__(self, expid):
         MainDataBase.__init__(self, expid)
@@ -563,7 +752,7 @@ class ExperimentStatus(MainDataBase):
         );''')
 
         if not os.path.exists(self.DB_FILE_AS_TIMES):
-            open(self.DB_FILE_AS_TIMES, "w")
+            # open(self.DB_FILE_AS_TIMES, "w")
             self.conn = self.create_connection(self.DB_FILE_AS_TIMES)
             self.create_table()
         else:
@@ -787,7 +976,7 @@ class JobDataStructure(MainDataBase):
                 job.job_name, None) and allJobsDict[job.job_name] == job.status] if current_job_data_last else None
             # Start processing
             if current_job_data:
-                print("There is current.")
+                # print("There is current.")
                 # Dropping parents key
                 for job in current_job_data:
                     if job.extra_data:
@@ -1082,13 +1271,13 @@ class JobDataStructure(MainDataBase):
         return result
 
     def get_historic_job_data(self, job_name):
-        """[summary]
+        """
+        Get the historic job data for a certain job  
 
-        Args:
-            job_name ([type]): [description]
-
-        Returns:
-            [type]: [description]
+        :param job_name: Name of Job  
+        :type job_name: str  
+        :return: JobData rows that match the job_name  
+        :rtype: list() of JobData objects
         """
         jobdata = []
         try:
@@ -1256,10 +1445,13 @@ class JobDataStructure(MainDataBase):
             return None
 
     def get_current_job_data(self, run_id, all_states=False):
-        """[summary]
-
-        Args:
-            run_id ([type]): [description]
+        """
+        Gets the job historical data for a run_id.  
+        :param run_id: Run identifier  
+        :type run_id: int  
+        :param all_states: False if only last=1 should be included, otherwise all rows  
+        :return: List of jobdata rows  
+        :rtype: list() of JobData objects
         """
         try:
             current_collection = []
@@ -1327,9 +1519,9 @@ class JobDataStructure(MainDataBase):
                             for run_id in set([job.run_id for job in current_collection])}
                     mean_sd = {run_id: (np.mean(data.get(run_id, [0])), np.std(data.get(run_id, [0])))
                                for run_id in set([job.run_id for job in current_collection])}
-                    threshold = 3
+                    threshold = 2
 
-                    return {run_id: [x for x in current_collection if x.run_id == run_id and (np.abs((x.running_time() - mean_sd.get(x.run_id, (0, 0))[0]) / mean_sd.get(x.run_id, (1, 1))[1]) if mean_sd.get(x.run_id, (1, 1))[1] > 0 else 0) <= threshold] for run_id in set([job.run_id for job in current_collection])}
+                    return {run_id: [x for x in current_collection if x.run_id == run_id and (np.abs((x.running_time() - mean_sd.get(x.run_id, (0, 0))[0]) / mean_sd.get(x.run_id, (1, 1))[1]) if mean_sd.get(x.run_id, (1, 1))[1] > 0 else 0) <= threshold and x.running_time() > 0] for run_id in set([job.run_id for job in current_collection])}
 
             return {}
         except Exception as exp:
@@ -1398,7 +1590,7 @@ class JobDataStructure(MainDataBase):
             if os.path.exists(self.folder_path):
                 current_job_data = self._get_current_job_data_last()
                 if current_job_data:
-                    print("Processing data")
+                    # print("Processing data")
                     time_0 = time.time()
                     # print(len(current_job_data))
                     #i = 0
@@ -1414,8 +1606,8 @@ class JobDataStructure(MainDataBase):
                             # If energy, then ignore extra_data
                             current_collection.append(JobData(jobitem.id, jobitem.counter, jobitem.job_name, jobitem.created, jobitem.modified, jobitem.submit, jobitem.start, jobitem.finish, jobitem.status, jobitem.rowtype, jobitem.ncpus,
                                                               jobitem.wallclock, jobitem.qos, jobitem.energy, jobitem.date, jobitem.section, jobitem.member, jobitem.chunk, jobitem.last, jobitem.platform, jobitem.job_id, None if jobitem.energy else jobitem.extra_data, jobitem.nnodes, jobitem.run_id))
-                    print("Processing data time {0}".format(
-                        time.time() - time_0))
+                    # print("Processing data time {0}".format(
+                    #     time.time() - time_0))
                     return current_collection
             return None
         except Exception as exp:
@@ -1474,7 +1666,11 @@ class JobDataStructure(MainDataBase):
             return None
 
     def _get_experiment_run_by_id(self, run_id):
-        """[summary]
+        """
+        :param run_id: Run Identifier  
+        :type run_id: int  
+        :return: First row that matches the run_id  
+        :rtype: Row as Tuple
         """
         try:
             if self.conn:
@@ -1500,6 +1696,33 @@ class JobDataStructure(MainDataBase):
                 run_id, "_get_experiment_run_by_id"))
             return None
 
+    # def _get_experiment_run_ids_from_period(self, period_start):
+    #     """
+    #     This function receives period start in timestamp format (int)  
+    #     Then it returns the unique run_id of the jobs that fall in this period.  
+    #     :param period_start: timestamp  
+    #     :type period_start: int   
+    #     :return: 
+    #     :rtype: list of int
+    #     """
+    #     try:
+    #         if self.conn:
+    #             self.conn.text_factory = str
+    #             cur = self.conn.cursor()
+    #             cur.execute(
+    #                 "select run_id from job_data where start >= ? or finish >= ? GROUP BY run_id", (period_start, period_start))
+    #             rows = cur.fetchall()
+    #             print(type(rows))
+    #             if len(rows) > 0:
+    #                 return rows
+    #             else:
+    #                 return None
+    #     except Exception as exp:
+    #         print(exp)
+    #         return None
+
+    #     return None
+
     def _select_pragma_version(self):
         """ Retrieves user_version from database
         """
@@ -1558,10 +1781,12 @@ class JobDataStructure(MainDataBase):
             return None
 
     def _get_current_job_data(self, run_id, all_states=False):
-        """Returns detail job_data by run_id
-
-        Args:
-            run_id ([type]): [description]
+        """
+        Get JobData by run_id.  
+        :param run_id: Run Identifier  
+        :type run_id: int  
+        :param all_states: False if only last=1, True all  
+        :type all_states: bool  
         """
         try:
             if self.conn:
@@ -1575,13 +1800,12 @@ class JobDataStructure(MainDataBase):
                         request_string = "SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id, MaxRSS, AveRSS, out, err, rowstatus  from job_data WHERE run_id=? and last=1 and finish > 0 and rowtype >= 2 ORDER BY id"
                     else:
                         request_string = "SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id from job_data WHERE run_id=? and last=1 and finish > 0 and rowtype >= 2 ORDER BY id"
-                    # cur.execute("SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id from job_data WHERE run_id=? and last=1 and finish > 0 and rowtype >= 2 ORDER BY id", (run_id,))
+
                 else:
                     if self.db_version == CURRENT_DB_VERSION:
                         request_string = "SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id, MaxRSS, AveRSS, out, err, rowstatus  from job_data WHERE run_id=? and rowtype >= 2 ORDER BY id"
                     else:
                         request_string = "SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id from job_data WHERE run_id=? and rowtype >= 2 ORDER BY id"
-                    # cur.execute("SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id from job_data WHERE run_id=? and rowtype >= 2 ORDER BY id", (run_id,))
 
                 cur.execute(request_string, (run_id,))
                 rows = cur.fetchall()
@@ -1671,7 +1895,7 @@ class JobDataStructure(MainDataBase):
     def _get_current_job_data_last(self):
         try:
             if self.conn:
-                print("Getting historic data.")
+                # print("Getting historic data.")
                 self.conn.text_factory = str
                 cur = self.conn.cursor()
                 request_string = "SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id from job_data WHERE last=1 and finish > 0 and rowtype >= 2 ORDER BY id"
@@ -1679,7 +1903,7 @@ class JobDataStructure(MainDataBase):
                     request_string = "SELECT id, counter, job_name, created, modified, submit, start, finish, status, rowtype, ncpus, wallclock, qos, energy, date, section, member, chunk, last, platform, job_id, extra_data, nnodes, run_id, MaxRSS, AveRSS, out, err, rowstatus from job_data WHERE last=1 and finish > 0 and rowtype >= 2 ORDER BY id"
                 cur.execute(request_string)
                 rows = cur.fetchall()
-                print("Retrieved historic.")
+                # print("Retrieved historic.")
                 return rows
         except sqlite3.Error as e:
             if _debug == True:
@@ -1758,11 +1982,12 @@ class JobDataStructure(MainDataBase):
                 # self.conn.commit()
                 return True
             return None
-        except sqlite3.Error as e:
+        except Exception as exp:
             if _debug == True:
                 print(traceback.format_exc())
-            print(traceback.format_exc())
-            print("Error on Insert : {}".format(str(type(e).__name__)))
+            # print(traceback.format_exc())
+            print("Error on _update_many_job_data : {}".format(
+                str(exp)))
             return None
 
 
